/*
 * Common system pager functionality.
 *
 * Copyright (C) 2022, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/re/c/rm.h>

#include <mem/memory_utils.h>

#include "pager.h"



/* Initialise common functionality. */

ExecPager::ExecPager(address_t start, address_t end)
: _start(start), _end(end)
{
}

/* Necessary virtual destructor. */

ExecPager::~ExecPager()
{
}



/* Add a region to the pager. */

void ExecPager::add(MappedRegion &region)
{
  _regions[region.start] = region;
}

/* Remove a region from the pager. */

void ExecPager::remove(MappedRegion &region)
{
  _regions.erase(region.start);
}



/* Attach a region for provision when page faults occur. */

long ExecPager::find(address_t *start, address_t *size, map_flags_t flags,
                     address_t offset, unsigned char align)
{
  if (align < L4_PAGESHIFT)
    align = L4_PAGESHIFT;

  /* Obtain the alignment increment and a properly aligned size. */

  address_t increment = 1UL << align;
  address_t region_size = round(*size, increment);

  /* Either attempt to find an address for the specified region, starting from
     any indicated address. */

  if (flags & L4RE_RM_F_SEARCH_ADDR)
  {
    /* Align any desired location. */

    address_t region_start = trunc(*start, increment);

    /* Enforce a minimum address. */

    if (region_start < _start)
      region_start = round(_start, increment);

    /* Search for existing regions after the desired, conformant address. */

    MappedRegions::iterator it = _regions.upper_bound(region_start);

    /* Consider any region preceding or encompassing the desired address. */

    if (it != _regions.begin())
    {
      MappedRegions::iterator next = it;

      /* Step back to the preceding region to get its details. */

      it--;
      MappedRegion &pr = it->second;
      address_t previous_end = pr.start + pr.size;
      it = next;

      /* The preceding region may displace the desired region location if it
         is encompassed by the region. */

      if (region_start < previous_end)
        region_start = round(previous_end, increment);
    }

    /* Consider the regions following the current region start candidate. */

    while ((it != _regions.end()) && !(_end && ((region_start + region_size) > _end)))
    {
      MappedRegion &r = it->second;

      /* Obtain the limit of available space being considered. */

      address_t end_limit = r.start;

      /* Investigate subsequent regions if not enough space exists between the
         preceding region (or start of memory) and the current region. */

      if ((region_start + region_size) > end_limit)
      {
        region_start = round(r.start + r.size, increment);
        it++;
      }

      /* Otherwise, the region can be positioned. */

      else
        break;
    }

    /* Test for enough memory constrained by any predefined limit. */

    if (_end && ((region_start + region_size) > _end))
      return -L4_ENOMEM;

    /* Return the configured start and size. */

    *start = region_start;
    *size = region_size;
    return L4_EOK;
  }

  /* Or attempt to add the specified region at a specific address. */

  else
  {
    // NOTE: To be implemented.

    return -L4_ENOMEM;
  }
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
